class EMA:
    def __init__(self, enable, model, decay=0.999):
        self.enable = enable
        self.model = model
        self.decay = decay
        self.shadow = {}
        self.backup = {}
        self.register()

    def register(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.shadow[name] = param.data.clone()

    def update(self):
        if self.enable:
            for name, param in self.model.named_parameters():
                if param.requires_grad:
                    assert name in self.shadow
                    average = (1.0 - self.decay) * param.data + self.decay * self.shadow[name]
                    self.shadow[name] = average.clone()

    def apply_shadow(self):
        if self.enable:
            for name, param in self.model.named_parameters():
                if param.requires_grad:
                    assert name in self.shadow
                    self.backup[name] = param.data
                    param.data = self.shadow[name]

    def restore(self):
        if self.enable:
            for name, param in self.model.named_parameters():
                if param.requires_grad:
                    assert name in self.backup
                    param.data = self.backup[name]
            self.backup = {}




import inspect
from .. import models


class Model:
    def __init__(self, config):
        model_name = config["name"]
        model_config = config[model_name]
        self.model = None

        if hasattr(models, model_name):
            model = getattr(models, model_name)
            if isinstance(model, type):
                sig = inspect.signature(model.__init__)
                params = {k: v for k, v in model_config.items() if k in sig.parameters}
                self.model = model(**params)

    def __call__(self):
        assert (self.model is not None)
        return self.model



import logging
import dill as pickle
from urllib import request


class Config:
    logging.basicConfig(level=logging.WARNING, format='%(message)s')

    def __init__(self, server_url, max_number):
        self.logger = logging.getLogger()
        self.server_url = server_url
        self.max_number = max_number
        self.config = self.search()

    def search(self):
        for i in range(self.max_number):
            try:
                response = request.urlopen(self.server_url)
                return pickle.load(response)
            except Exception as e:
                self.logger.warning(f'load datasets config failed, restart...')
        self.logger.warning('try timeout, exceed the max_number, terminal!')

    def get_config(self):
        return self.config



from tqdm import tqdm
from ..utils import Metric
from ..utils import gather


class BaseTrainer:
    def __init__(self, train_loader, model, criterion, optimizer, scheduler, ema, device):
        self.device = device
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.train_loader = train_loader
        self.train_size = len(train_loader.dataset)
        self.steps = len(train_loader)
        self.lr = self.scheduler.get_last_lr()[0]
        self.ema = ema

    def __call__(self, epoch, logger, header, master, world_size):
        self.model.train()
        metric = Metric(self.steps)
        process = tqdm(total=self.steps, desc=" " * 72, disable=not master)
        self.train_loader.sampler.set_epoch(epoch)
        for step, (images, labels, _) in enumerate(self.train_loader):
            images = images.to(self.device)
            labels = labels.float().squeeze().to(self.device)
            self.optimizer.zero_grad()
            outputs = self.model(images)
            loss = self.criterion(outputs, labels)
            loss.backward()
            self.optimizer.step()
            self.scheduler.step()
            self.ema.update()
            self.lr = self.scheduler.get_last_lr()[0]
            loss, outputs, labels = gather(loss, outputs, labels, world_size)
            loss, precision, recall, f1 = metric(loss, outputs, labels)
            if step < self.steps - 1:
                logger(header + [self.lr, loss, precision, recall, f1], process)
        m_loss, m_precision, m_recall, m_f1 = metric.release()
        logger(header + [self.lr, m_loss, m_precision, m_recall, m_f1], process)
        return m_loss, m_precision, m_recall, m_f1



import cv2
import lmdb
import numpy as np
from .transform import Transform
from torch.utils.data.dataset import Dataset
from ..utils import Crypt


class LMDBDataset(Dataset):
    def __init__(self, config, mode=None):
        self.mode = mode
        self.config = config

        self.context = None
        self.lmdb_path = config["lmdb_path"]
        self.header = self.get_header()
        self.description = self.get_description()
        self.samples = self.get_samples()
        self.transform = self.get_transform()
        self.nc = len(self.header)
        self.crypt = Crypt(self.config["key"])

    def __getitem__(self, index):
        self.init_context()
        key = self.samples[index]
        sample = self.crypt.decode(key)
        name, index, _ = eval(sample)
        label = np.zeros(self.nc, np.float64)
        label[index] = 1.0
        buffer = self.context.get(key)
        buffer = np.frombuffer(buffer, dtype=np.uint8)
        image = cv2.imdecode(buffer, cv2.IMREAD_COLOR)
        image, label = self.transform(image, label)
        return image, label, name

    def __len__(self):
        return len(self.samples)

    def init_context(self):
        if self.context is None:
            self.context = self.get_context()

    def get_context(self):
        lmdb_file = lmdb.open(self.lmdb_path)
        return lmdb_file.begin()

    def get_header(self):
        context = self.get_context()
        buffer = context.get("header".encode("utf-8"))
        return eval(buffer.decode("utf-8"))

    def get_description(self):
        context = self.get_context()
        buffer = context.get("description".encode("utf-8"))
        return eval(buffer.decode("utf-8"))

    def get_samples(self):
        assert (self.mode is not None)
        context = self.get_context()
        buffer = context.get(self.mode.encode("utf-8"))
        return eval(buffer.decode("utf-8"))

    def get_transform(self):
        assert (self.mode is not None)
        config_transform = self.config["transform"][self.mode]
        return Transform(config_transform, self.header)



import inspect
from torch import optim


class Optimizer:
    def __init__(self, config, model):
        optim_name = config["name"]
        optim_config = config[optim_name]
        self.optimizer = None
        if hasattr(optim, optim_name):
            optimizer = getattr(optim, optim_name)
            if isinstance(optimizer, type):
                sig = inspect.signature(optimizer.__init__)
                params = {k: v for k, v in optim_config.items() if k in sig.parameters}
                self.optimizer = optimizer(params=model.parameters(), **params)

    def __call__(self):
        assert (self.optimizer is not None)
        return self.optimizer



import torch
import torch.distributed as dist
from torch.nn.parallel import DataParallel, DistributedDataParallel


def gather(loss, outputs, labels, world_size):
    all_loss, all_outputs, all_labels = [], [], []
    for _ in range(world_size):
        all_loss.append(torch.empty_like(loss.data.unsqueeze(0)))
        all_outputs.append(torch.empty_like(outputs.data))
        all_labels.append(torch.empty_like(labels))
    dist.all_gather(all_loss, loss.data.unsqueeze(0))
    dist.all_gather(all_outputs, outputs.data)
    dist.all_gather(all_labels, labels)
    loss = torch.cat(all_loss, dim=0).sum().item()
    outputs = torch.cat(all_outputs, dim=0)
    labels = torch.cat(all_labels, dim=0)
    return loss, outputs, labels


def de_parallel(model):
    if isinstance(model, (DataParallel, DistributedDataParallel)):
        model = model.module
    return model



import inspect
from torch.optim import lr_scheduler


class Scheduler:
    def __init__(self, config, optimizer):
        scheduler_name = config["name"]
        scheduler_config = config[scheduler_name]
        self.scheduler = None
        if hasattr(lr_scheduler, scheduler_name):
            scheduler = getattr(lr_scheduler, scheduler_name)
            if isinstance(scheduler, type):
                sig = inspect.signature(scheduler.__init__)
                params = {k: v for k, v in scheduler_config.items() if k in sig.parameters}
                self.scheduler = scheduler(optimizer=optimizer, **params)

    def __call__(self):
        assert (self.scheduler is not None)
        return self.scheduler



import cv2
import random
import numpy as np
import albumentations as a
import torch
import torch.nn.functional as func
from torchvision import transforms


class LetterBox:
    def __init__(self, config, header):
        self.p = config["p"]
        self.fill = config["fill"]

    def __call__(self, data):
        image, label = data
        if random.random() < self.p:
            c, h, w = image.shape
            size = max(w, h)
            pad_v = (size - h) // 2
            pad_h = (size - w) // 2
            pad = (pad_h, pad_h, pad_v, pad_v)
            pad_img = func.pad(image, pad, mode='constant', value=self.fill)
        else:
            pad_img = image
        output = (pad_img, label)
        return output


class ToTensor:
    def __init__(self, config, header):
        self.transform = transforms.ToTensor()

    def __call__(self, data):
        image, label = data
        tensor = self.transform(image)
        output = (tensor, label)
        return output


class Resize:
    def __init__(self, config, header):
        width, height = config["size"]
        self.transform = transforms.Resize([height, width])

    def __call__(self, data):
        image, label = data
        resized_img = self.transform(image)
        output = (resized_img, label)
        return output


class RandomRotation:
    def __init__(self, config, header):
        p = config["p"]
        fill = config["fill"]
        degrees = config["degrees"]
        transform = transforms.RandomRotation(degrees, fill=fill)
        self.transform = transforms.RandomApply([transform], p=p)

    def __call__(self, data):
        image, label = data
        rotate_img = self.transform(image)
        output = (rotate_img, label)
        return output


class RandomHorizontalFlip:
    def __init__(self, config, header):
        self.p = config["p"]
        self.is_swap = False
        if "swap" in config.keys():
            head = config.pop("swap")
            self.swap = [header.index(h) for h in head]
            self.is_swap = len(head) == 2
            assert self.is_swap

    def __call__(self, data):
        image, label = data
        if random.random() < self.p:
            flipped_img = torch.flip(image, dims=[2])
            if self.is_swap:
                for index in self.swap:
                    label[index] = label[index] ^ 1
        else:
            flipped_img = image
        output = (flipped_img, label)
        return output


class RandomErasing:
    """
    "RandomErasing": {
          "p": 0.3,
          "scale": [
            0.02,
            0.125
          ],
          "ratio": [
            0.33,
            3.0
          ],
          "value": 0.5,
          "inplace": false
        },
    """

    def __init__(self, config, header):
        p = config["p"]
        scale = tuple(config["scale"])
        ratio = tuple(config["ratio"])
        value = config["value"]
        inplace = config["inplace"]
        self.transform = transforms.RandomErasing(p=p,
                                                  scale=scale,
                                                  ratio=ratio,
                                                  value=value,
                                                  inplace=inplace)

    def __call__(self, data):
        image, label = data
        erase_img = self.transform(image)
        output = (erase_img, label)
        return output


class RandomCrop:
    def __init__(self, config, header):
        self.p = config["p"]
        height, width = config["size"]
        fill = config["fill"]
        self.transform_crop = transforms.RandomCrop(size=[height, width], fill=fill)
        self.transform_resize = transforms.Resize([height, width])

    def __call__(self, data):
        image, label = data
        if random.random() < self.p:
            crop_img = self.transform_crop(image)
        else:
            crop_img = self.transform_resize(image)
        output = (crop_img, label)
        return output


class ColorJitter:
    def __init__(self, config, header):
        p = config["p"]
        brightness = config["brightness"]
        contrast = config["contrast"]
        saturation = config["saturation"]
        hue = config["hue"]
        transform = transforms.ColorJitter(brightness=brightness,
                                           contrast=contrast,
                                           saturation=saturation,
                                           hue=hue)
        self.transform = transforms.RandomApply([transform], p=p)

    def __call__(self, data):
        image, label = data
        color_jitter_img = self.transform(image)
        output = (color_jitter_img, label)
        return output


class Normalize:
    def __init__(self, config, header):
        self.p = config["p"]
        mean = tuple(config["mean"])
        std = tuple(config["std"])
        inplace = config["inplace"]
        self.transform = transforms.Normalize(mean=mean,
                                              std=std,
                                              inplace=inplace)

    def __call__(self, data):
        image, label = data
        if random.random() < self.p:
            normalized_img = self.transform(image)
        else:
            normalized_img = image
        output = (normalized_img, label)
        return output


class GrayScale:
    def __init__(self, config, header):
        p = config["p"]
        self.transform = transforms.RandomGrayscale(p=p)

    def __call__(self, data):
        image, label = data
        gray_img = self.transform(image)
        output = (gray_img, label)
        return output


class Blur:
    def __init__(self, config, header):
        self.p = config["p"]
        transform_list = [
            a.Blur(p=self.p),
            a.GaussianBlur(p=self.p),
            a.MedianBlur(p=self.p),
            a.CLAHE(p=self.p),
            a.RandomGamma(p=self.p),
            a.ImageCompression(p=self.p * 2,
                               quality_lower=75),
            a.RandomFog(p=self.p),
            a.RandomRain(p=self.p),
            a.RandomSnow(p=self.p),
            a.RandomShadow(p=self.p),
        ]
        self.transform = a.Compose(transform_list, p=1.0)

    def __call__(self, data):
        image, label = data
        blur_img = self.transform(image=image)['image']
        output = (blur_img, label)
        return output


class GridDropout:
    """
    "GridDropout": {
          "p": 0.99,
          "max_ratio": 0.5,
          "unit_size_min": 5,
          "unit_size_max": 30
        },
    """

    def __init__(self, config, header):
        self.p = config["p"]
        self.max_ratio = config["max_ratio"]
        self.unit_size_min = config["unit_size_min"]
        self.unit_size_max = config["unit_size_max"]

    def __call__(self, data):
        image, label = data
        ratio = self.max_ratio * random.random()
        fill_value = random.randint(0, 255)
        dropout = a.GridDropout(p=self.p, ratio=ratio,
                                random_offset=True,
                                holes_number_x=5,
                                holes_number_y=10,
                                unit_size_min=self.unit_size_min,
                                unit_size_max=self.unit_size_max,
                                fill_value=fill_value)
        dropout_img = dropout(image=image)['image']
        output = (dropout_img, label)
        return output


class GridMix:
    def __init__(self, config, header):
        self.p = config["p"]
        self.max_ratio = config["max_ratio"]
        self.min_divide = config["min_divide"]
        self.max_divide = config["max_divide"]

    def __call__(self, data):
        if random.random() < self.p:
            image, label = data
            ratio = random.uniform(0, self.max_ratio)
            x_grid_number = random.randint(self.min_divide, self.max_divide)
            y_grid_number = random.randint(self.min_divide, self.max_divide)

            height, width, _ = image.shape
            grid_width = width // x_grid_number
            grid_height = height // y_grid_number
            grid_image = image.copy()
            for _ in range(int(x_grid_number * y_grid_number * ratio)):
                row = np.random.randint(0, y_grid_number)
                col = np.random.randint(0, x_grid_number)
                start_row = row * grid_height
                end_row = start_row + grid_height
                start_col = col * grid_width
                end_col = start_col + grid_width
                if random.random() > 0.5:
                    fill = np.random.randint(0, 128, size=3)
                else:
                    fill = np.random.randint(0, 256)
                grid_image[start_row:end_row, start_col:end_col] = fill
            output = (grid_image, label)
        else:
            output = data
        return output


class AutoMix:
    def __init__(self, config, header):
        self.p = config["p"]
        self.min_ratio, self.max_ratio = config["ratio"]
        self.min_divide, self.max_divide = config["divide"]
        self.attach_image = None
        self.attach_label = None

    def __call__(self, data):
        if (random.random() < self.p
                and self.attach_image is not None
                and self.attach_label is not None):
            source_image, source_label = data
            height, width, _ = source_image.shape
            attach_image = cv2.resize(self.attach_image, (width, height))
            ratio = random.uniform(self.min_ratio, self.max_ratio)
            x_grid_number = random.randint(self.min_divide, self.max_divide)
            y_grid_number = random.randint(self.min_divide, self.max_divide)

            grid_width = width // x_grid_number
            grid_height = height // y_grid_number
            mixed_image = source_image.copy()
            for _ in range(int(x_grid_number * y_grid_number * ratio)):
                row = np.random.randint(0, y_grid_number)
                col = np.random.randint(0, x_grid_number)
                y1 = row * grid_height
                y2 = y1 + grid_height
                x1 = col * grid_width
                x2 = x1 + grid_width
                mixed_image[y1:y2, x1:x2] = attach_image[y1:y2, x1:x2]
            mixed_label = source_label * (1 - ratio) + self.attach_label * ratio
            self.attach_image, self.attach_label = data
            output = (mixed_image, mixed_label)
        else:
            self.attach_image, self.attach_label = data
            output = data
        return output


class MixUP:
    def __init__(self, config, header):
        self.p = config["p"]
        self.min_alpha, self.max_alpha = config["alpha"]
        self.attach_image = None
        self.attach_label = None

    def __call__(self, data):
        if (random.random() < self.p
                and self.attach_image is not None
                and self.attach_label is not None):
            source_image, source_label = data
            alpha = random.uniform(self.min_alpha, self.max_alpha)
            mixed_image = torch.add(torch.mul(source_image, alpha),
                                    torch.mul(self.attach_image, 1 - alpha))
            mixed_label = source_label * alpha + self.attach_label * (1 - alpha)
            self.attach_image, self.attach_label = data
            output = (mixed_image, mixed_label)
        else:
            output = data
            self.attach_image, self.attach_label = data
        return output



import torch
import torch.nn as nn


def init_weights(m):
    if isinstance(m, (nn.Conv2d, nn.Linear)):
        torch.nn.init.xavier_normal_(m.weight)
        if m.bias is not None:
            m.bias.data.fill_(0.01)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels,
                               kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels,
                               kernel_size=3, stride=1,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != self.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, self.expansion * out_channels,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * out_channels))

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = self.relu(out)
        return out


class ResNet18(nn.Module):
    def __init__(self, num_classes=1000, depth=1.0, width=1.0, weights=""):
        super(ResNet18, self).__init__()
        self.in_channels = round(64 * width)
        self.backbone = nn.Sequential(
            nn.Conv2d(3, round(64 * width), kernel_size=7, stride=2, padding=3, bias=False),
            nn.BatchNorm2d(round(64 * width)),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            self.make_layer(BasicBlock, round(64 * width), round(2 * depth), stride=1),
            self.make_layer(BasicBlock, round(128 * width), round(2 * depth), stride=2),
            self.make_layer(BasicBlock, round(256 * width), round(2 * depth), stride=2),
            self.make_layer(BasicBlock, round(512 * width), round(2 * depth), stride=2),
            nn.AdaptiveAvgPool2d((1, 1))
        )
        self.head = nn.Sequential(
            nn.Linear(round(512 * width) * BasicBlock.expansion, num_classes),
            nn.Sigmoid()
        )
        self.init_weights(weights)

    def init_weights(self, weights):
        if len(weights):
            backbone = torch.load(weights)["model"].backbone
            self.backbone.load_state_dict(backbone.state_dict())
            self.head.apply(init_weights)
        else:
            self.apply(init_weights)

    def make_layer(self, block, out_channels, blocks, stride):
        layers = [block(self.in_channels, out_channels, stride)]
        self.in_channels = out_channels * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.in_channels, out_channels))
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.backbone(x)
        x = torch.flatten(x, 1)
        x = self.head(x)
        return x




import torch
import torch.multiprocessing as mp
from .context import Context


def run(rank, world_size, config):
    config["rank"] = rank
    config["world_size"] = world_size
    trainer = Context(config)
    trainer.train()


def main(config):
    world_size = len(config["train"]["device"])
    device_num = torch.cuda.device_count()
    assert world_size <= device_num
    mp.spawn(run, args=(world_size, config), nprocs=world_size, join=True)



from cryptography.fernet import Fernet


class Crypt:
    def __init__(self, key):
        self.key = key
        self.crypt = Fernet(key) if len(key) else None

    def encode(self, data):
        if self.crypt is None:
            data = data.encode("utf-8")
        else:
            data = self.crypt.encrypt(data.encode("utf-8"))
        return data

    def decode(self, data):
        if self.crypt is None:
            data = data.decode("utf-8")
        else:
            data = self.crypt.decrypt(data).decode("utf-8")
        return data



import math
import logging
from tqdm import tqdm


class Logger:
    logging.basicConfig(level=logging.WARNING, format='%(message)s')

    def __init__(self, enable=True, space=12, scale=1e-6):
        self.enable = enable
        self.space = space
        self.scale = int(abs(math.log10(scale)))

    def __call__(self, params, pbar=None):
        if self.enable:
            if isinstance(params, list):
                description = self.merge(params)
                if isinstance(pbar, tqdm):
                    pbar.set_description(description)
                    pbar.update()
                else:
                    logging.warning(description)
            elif isinstance(params, str):
                logging.warning(params)
            else:
                logging.warning("This type of printing is not supported !")

    def merge(self, params):
        if isinstance(params, list):
            return "".join(list(map(self.format, params)))
        else:
            return params

    def format(self, param):
        if isinstance(param, float):
            return f"{f'{param:.{self.scale}f}':>{self.space}}"
        else:
            return f"{f'{param}':>{self.space}}"



import onnx
import onnxsim
import torch
from torchsummary import summary

pth_path = "../../runs/pretrain/botnet/weights/epoch120.pt"
onnx_path = "../../runs/pretrain/botnet/weights/epoch120.onnx"

model = torch.load(pth_path)["model"].to("cuda")
summary(model, batch_size=1, input_size=(3, 224, 224), device="cuda")
model.eval()

data = torch.rand((1, 3, 224, 224), dtype=torch.float32, device="cuda")
torch.onnx.export(model, data, onnx_path, verbose=False, opset_version=12,
                  input_names=["images"], output_names=["output"],
                  # dynamic_axes={
                  #     'images': {0: 'batch_size'},
                  #     'output': {0: 'batch_size'}}
                  )

# Checks
model_onnx = onnx.load(onnx_path)  # load onnx model
onnx.checker.check_model(model_onnx)  # check onnx model

# Simplify
try:
    print(f'simplifying with onnxsim {onnxsim.__version__}...')
    model_onnx, check = onnxsim.simplify(model_onnx)
    assert check, 'Simplified ONNX model could not be validated'
except Exception as e:
    print(f'simplifier failure: {e}')

onnx.save(model_onnx, onnx_path)



import os
import re
import copy
import pandas as pd
import dill as pickle
from io import BytesIO

import torch
from torch.cuda import amp
from torch import nn, optim
from torch import distributed as dist
from torch.nn.parallel import DistributedDataParallel
from torchsummary import summary

from .models import Model
from .datasets import LMDBDataset, LMDBDataLoader
from .engine import BaseTrainer, BaseValidator
from .utils import Loss, Optimizer, Scheduler, EMA
from .utils import Storager, Logger, Event
from .utils import de_parallel

from . import __version__


class Context:
    def __init__(self, config):
        self.rank = config["rank"]
        self.world_size = config["world_size"]
        self.addr = config["network"]["addr"]
        self.port = config["network"]["port"]
        self.master_process = self.rank == 0 if self.world_size >= 1 else self.rank == -1
        self.logger = Logger(self.master_process)

        self.config = config
        self.config_dataset = self.config["dataset"]
        self.config_dataloader = self.config["dataloader"]
        self.config_dataloader["rank"] = self.rank
        self.config_dataloader["world_size"] = self.world_size
        self.config_model = self.config["model"]
        self.config_train = self.config["train"]

        self.epoch = 0
        self.epochs = self.config_train["epochs"]
        self.save_dir = self.config_train["save_dir"]
        self.save_period = self.config_train["save_period"]
        self.device = self.config_train["device"][self.rank]
        self.resume = self.config_train["resume"]

        self.save_path = self.check_storage_path()
        self.storager, self.event = self.setup_train()
        self.logger("The results are saved in: %s" % self.save_path)

        self.train_dataset = LMDBDataset(self.config_dataset, "train")
        self.valid_dataset = LMDBDataset(self.config_dataset, "valid")
        self.train_loader = LMDBDataLoader(self.config_dataloader, self.train_dataset, "train")
        self.valid_loader = LMDBDataLoader(self.config_dataloader, self.valid_dataset, "valid")
        self.logger("%s size: %d" % ("train", len(self.train_dataset)))
        self.logger("%s size: %d" % ("valid", len(self.valid_dataset)))

        self.model = Model(self.config_model)().to(self.device)
        self.criterion = Loss(self.config_train["loss"])().to(self.device)
        self.optimizer = Optimizer(self.config_train["optimizer"], self.model)()
        self.scheduler = Scheduler(self.config_train["scheduler"], self.optimizer)()
        self.ema = EMA(self.config_train["ema"], self.model)

        self.resume_training()
        self.setup_ddp()
        self.setup_model()

        self.trainer = BaseTrainer(
            train_loader=self.train_loader,
            model=self.model,
            criterion=self.criterion,
            optimizer=self.optimizer,
            scheduler=self.scheduler,
            ema=self.ema,
            device=self.device)
        self.validator = BaseValidator(
            valid_loader=self.valid_loader,
            model=self.model,
            criterion=self.criterion,
            ema=self.ema,
            device=self.device)

    def __del__(self):
        dist.destroy_process_group()

    def train(self):
        while self.epoch < self.epochs:
            self.logger(["Epoch", "LR", "Loss", "P", "R", "F1"])
            self.epoch = self.epoch + 1
            train_result = self.trainer(
                logger=self.logger,
                epoch=self.epoch,
                header=[f"{self.epoch}/{self.epochs}"],
                master=self.master_process,
                world_size=self.world_size)
            valid_result = self.validator(
                epoch=self.epoch,
                logger=self.logger,
                header=["", ""],
                master=self.master_process,
                world_size=self.world_size)
            if self.master_process:
                self.save_model()
                lr = self.scheduler.get_last_lr()[0]
                self.event(epoch=self.epoch,
                           learning_rate=lr,
                           train_result=train_result,
                           valid_result=valid_result)
        self.epoch = 0

    def setup_ddp(self):
        os.environ["MASTER_ADDR"] = self.addr
        os.environ["MASTER_PORT"] = self.port
        backend = 'nccl' if dist.is_nccl_available() else 'gloo'
        dist.init_process_group(backend=backend, rank=self.rank,
                                world_size=self.world_size)
        torch.cuda.set_device(self.device)

    def setup_model(self):
        if self.master_process:
            summary(self.model, tuple(self.config_model["shape"]), -1, "cuda")
        if self.world_size > 1:
            self.model = DistributedDataParallel(
                module=self.model,
                device_ids=[self.device])

    def setup_train(self):
        if self.master_process:
            storager = Storager(self.save_path, self.config_dataset["key"])
            event = Event(storager)
        elif self.resume:
            storager = Storager(self.save_path, self.config_dataset["key"])
            event = None
        else:
            storager, event = None, None
        return storager, event

    def save_model(self):
        """Save model checkpoints based on various conditions."""
        if self.epoch % self.save_period == 0:
            buffer = BytesIO()
            ckpt = {
                'epoch': self.epoch,
                'model': de_parallel(copy.deepcopy(self.model)),
                'ema': de_parallel(copy.deepcopy(self.ema.shadow)),
                'optimizer': pickle.dumps(self.optimizer.state_dict()),
                'scheduler': pickle.dumps(self.scheduler.state_dict()),
                'config': self.config,
                'version': __version__
            }
            torch.save(ckpt, buffer, pickle_module=pickle)
            model_path = f"./weights/epoch{self.epoch}.pt"
            self.storager.write(model_path, buffer)

    def check_storage_path(self):
        current_id = 0
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)
        for file_name in os.listdir(self.save_dir):
            match = re.search(r'exp(\d+)\.lmdb', file_name)
            if match is not None:
                match_id = int(match.group(1))
                if match_id > current_id:
                    current_id = match_id
        self.resume = True if self.resume and current_id != 0 else False
        current_id = current_id if self.resume else current_id + 1
        save_path = os.path.join(self.save_dir, "exp{}.lmdb".format(current_id))
        return save_path

    def resume_training(self):
        """Resume training from given epoch."""
        if self.resume:
            csv_key = "./result.csv"
            csv_byte = self.storager.read(csv_key)
            if csv_byte is not None:
                csv_buffer = BytesIO(csv_byte)
                result = pd.read_csv(csv_buffer)
                if self.master_process:
                    self.event.reload(result)
                self.epoch = len(result)

                pt_key = f"./weights/epoch{self.epoch}.pt"
                pt_byte = self.storager.read(pt_key)
                pt_buffer = BytesIO(pt_byte)
                weight = torch.load(pt_buffer)

                state_dict = weight["model"].state_dict()
                self.model.load_state_dict(state_dict)
                self.optimizer.load_state_dict(pickle.loads(weight["optimizer"]))
                self.scheduler.load_state_dict(pickle.loads(weight["scheduler"]))
                self.ema.shadow = weight["ema"].copy()



import os
import time
import copy
import pandas as pd
from io import BytesIO
from subprocess import Popen, PIPE
from torch.utils.tensorboard import SummaryWriter


class Event:
    def __init__(self, storager):
        self.storager = storager
        self.writer = SummaryWriter(self.storager.save_path)
        self.process = self.start_tensorboard()
        self.metrics = {
            "epoch": [],
            "precision/train": [],
            "precision/valid": [],
            "recall/train": [],
            "recall/valid": [],
            "f1/train": [],
            "f1/valid": [],
            "loss/train": [],
            "loss/valid": [],
            "learning_rate": []
        }

    def __call__(self, epoch, learning_rate, train_result, valid_result):
        self.metrics["epoch"].append(epoch)
        self.metrics["learning_rate"].append(learning_rate)
        self.metrics["loss/train"].append(train_result[0])
        self.metrics["loss/valid"].append(valid_result[0])
        self.metrics["precision/train"].append(train_result[1])
        self.metrics["precision/valid"].append(valid_result[1])
        self.metrics["recall/train"].append(train_result[2])
        self.metrics["recall/valid"].append(valid_result[2])
        self.metrics["f1/train"].append(train_result[3])
        self.metrics["f1/valid"].append(valid_result[3])
        self.write_csv()
        self.write_summary()

    def __del__(self):
        self.writer.close()
        self.process.terminate()

    def start_tensorboard(self):
        """Start a tensorboard process"""
        cmd = ["tensorboard",
               "--logdir", self.storager.save_path,
               "--load_fast", "false",
               "--port", "6006"]
        process = Popen(cmd)
        time.sleep(5)
        return process

    def reload(self, result):
        metrics = result.to_dict()
        for key, value in metrics.items():
            self.metrics[key] = list(value.values())

        # concat tensorboard output
        # event_list = []
        # for event_name in os.listdir(self.storager.save_path):
        #     if event_name.startswith("events"):
        #         event_list.append(os.path.join(self.storager.save_path, event_name))
        # buffer = BytesIO()
        # for event_path in event_list:
        #     with open(event_path, "rb") as event_file:
        #         buffer.write(event_file.read())
        #         os.remove(event_path)
        # output_path = event_list[-1]
        # with open(output_path, "wb") as output_file:
        #     output_file.write(buffer.getvalue())

    def write_csv(self):
        """Saves training metrics to a CSV file."""
        buffer = BytesIO()
        df = pd.DataFrame(self.metrics)
        df.to_csv(buffer, index=False)
        csv_path = "./result.csv"
        self.storager.write(csv_path, buffer)

    def write_summary(self):
        """Saves training metrics to tensorboard."""
        metrics = copy.deepcopy(self.metrics)
        epoch = metrics.pop("epoch")[-1]
        for name, metric in metrics.items():
            self.writer.add_scalar(name, metric[-1], epoch)



import torch


class Metric:
    def __init__(self, size):
        self.size = size
        self.all_loss = 0.0
        self.all_output_num = 0.0
        self.all_label_num = 0.0
        self.all_right_num = 0.0
        self.ins = 1e-8

    def __call__(self, loss: float, outputs: torch.Tensor, labels: torch.Tensor):
        outputs = outputs.ge(0.5)
        labels = labels == 1.0
        output_num = outputs.sum().item()
        label_num = labels.sum().item()
        right_num = (outputs & labels).sum().item()
        self.all_loss += loss
        self.all_output_num += output_num
        self.all_label_num += label_num
        self.all_right_num += right_num
        precision = right_num / (output_num + self.ins)
        recall = right_num / (label_num + self.ins)
        f1 = 2 * (precision * recall) / (precision + recall + self.ins)
        return loss, precision, recall, f1

    def release(self):
        m_loss = self.all_loss / (self.size + self.ins)
        m_precision = self.all_right_num / (self.all_output_num + self.ins)
        m_recall = self.all_right_num / (self.all_label_num + self.ins)
        m_f1 = 2 * (m_precision * m_recall) / (m_precision + m_recall + self.ins)
        self.all_loss = 0.0
        self.all_output_num = 0.0
        self.all_label_num = 0.0
        self.all_right_num = 0.0
        return m_loss, m_precision, m_recall, m_f1



from tqdm import tqdm
from ..utils import Metric
from ..utils import gather


class BaseValidator:
    def __init__(self, valid_loader, model, criterion, ema, device):
        self.device = device
        self.model = model
        self.criterion = criterion
        self.valid_loader = valid_loader
        self.valid_size = len(valid_loader.dataset)
        self.steps = len(valid_loader)
        self.ema = ema

    def __call__(self, epoch, logger, header, master, world_size):
        self.model.eval()
        self.ema.apply_shadow()
        metric = Metric(self.steps)
        process = tqdm(total=self.steps, desc=" " * 72, disable=not master)
        self.valid_loader.sampler.set_epoch(epoch)
        for step, (images, labels, _) in enumerate(self.valid_loader):
            images = images.to(self.device)
            labels = labels.float().squeeze().to(self.device)
            outputs = self.model(images)
            loss = self.criterion(outputs, labels)
            loss, outputs, labels = gather(loss, outputs, labels, world_size)
            loss, precision, recall, f1 = metric(loss, outputs, labels)
            if step < self.steps - 1:
                logger(header + [loss, precision, recall, f1], process)
        self.ema.restore()
        m_loss, m_precision, m_recall, m_f1 = metric.release()
        logger(header + [m_loss, m_precision, m_recall, m_f1], process)
        return m_loss, m_precision, m_recall, m_f1



from .loss import Loss
from .metric import Metric
from .optimizer import Optimizer
from .scheduler import Scheduler
from .storager import Storager
from .logger import Logger
from .event import Event
from .ema import EMA
from .crypt import Crypt
from .parallel import gather, de_parallel

__all__ = ("Loss", "Metric", "Optimizer", "Scheduler",
           "Storager", "Logger", "Event", "EMA", "Crypt",
           "gather", "de_parallel")



import torch
import torch.nn as nn


class BotNet(nn.Module):
    def __init__(self, num_classes=1000, depth=1.0, width=1.0, weights=""):
        super(BotNet, self).__init__()

        self.in_channels = 3
        self.backbone = nn.Sequential(
            self.make_conv(Conv, out_channels=round(64 * width), repeats=1, k=3, s=2),
            self.make_conv(Conv, out_channels=round(128 * width), repeats=1, k=3, s=2),
            self.make_c2f(C2f, out_channels=round(128 * width), repeats=round(6 * depth), shortcut=True),
            self.make_conv(Conv, out_channels=round(256 * width), repeats=1, k=3, s=2),
            self.make_c2f(C2f, out_channels=round(256 * width), repeats=round(6 * depth), shortcut=True),
            self.make_conv(Conv, out_channels=round(512 * width), repeats=1, k=3, s=2),
            self.make_c2f(C2f, out_channels=round(512 * width), repeats=round(6 * depth), shortcut=True),
            self.make_conv(Conv, out_channels=round(1024 * width), repeats=1, k=3, s=2),
            self.make_c2f(C2f, out_channels=round(1024 * width), repeats=round(6 * depth), shortcut=True)
        )
        self.head = Head(self.in_channels, num_classes)

        self.init_weights(weights)

    def init_weights(self, weights):
        if len(weights):
            model = torch.load(weights)["model"]
            self.load_state_dict(model.state_dict())
        else:
            self.apply(init_weights)

    def make_conv(self, block, out_channels, repeats, k, s):
        layers = []
        for _ in range(0, repeats):
            layers.append(block(self.in_channels, out_channels, k=k, s=s))
            self.in_channels = out_channels
        return nn.Sequential(*layers)

    def make_c2f(self, block, out_channels, repeats, shortcut):
        layers = []
        for _ in range(0, repeats):
            layers.append(block(self.in_channels, out_channels, shortcut=shortcut))
            self.in_channels = out_channels
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.backbone(x)
        x = self.head(x)
        return x


def init_weights(m):
    if isinstance(m, (nn.Conv2d, nn.Linear)):
        torch.nn.init.xavier_normal_(m.weight)
        if m.bias is not None:
            m.bias.data.fill_(0.01)


def auto_pad(k, p=None, d=1):  # kernel, padding, dilation
    """Pad to 'same' shape outputs."""
    if d > 1:
        k = d * (k - 1) + 1 if isinstance(k, int) else [d * (x - 1) + 1 for x in k]  # actual kernel-size
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


class Conv(nn.Module):
    """Standard convolution with args(ch_in, ch_out, kernel, stride, padding, groups, dilation, activation)."""

    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, d=1, act=True):
        """Initialize Conv layer with given arguments including activation."""
        super().__init__()
        default_act = [nn.ReLU(), nn.SiLU()][0]
        self.conv = nn.Conv2d(c1, c2, k, s, auto_pad(k, p, d), groups=g, dilation=d, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = default_act if act is True else act if isinstance(act, nn.Module) else nn.Identity()

    def forward(self, x):
        """Apply convolution, batch normalization and activation to input tensor."""
        return self.act(self.bn(self.conv(x)))

    def forward_fuse(self, x):
        """Perform transposed convolution of 2D data."""
        return self.act(self.conv(x))


class C2f(nn.Module):
    """Faster Implementation of CSP Bottleneck with 2 convolutions."""

    def __init__(self, c1, c2, n=1, shortcut=False, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        self.c = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, 2 * self.c, 1, 1)
        self.cv2 = Conv((2 + n) * self.c, c2, 1, 1)  # optional act=FReLU(c2)
        self.m = nn.ModuleList(Bottleneck(self.c, self.c, shortcut, g, k=((3, 3), (3, 3)), e=1.0) for _ in range(n))

    def forward(self, x):
        """Forward pass through C2f layer."""
        y = list(self.cv1(x).chunk(2, 1))
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))

    def forward_split(self, x):
        """Forward pass using split() instead of chunk()."""
        y = list(self.cv1(x).split((self.c, self.c), 1))
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))


class Bottleneck(nn.Module):
    """Standard bottleneck."""

    def __init__(self, c1, c2, shortcut=True, g=1, k=(3, 3), e=0.5):  # ch_in, ch_out, shortcut, groups, kernels, expand
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, k[0], 1)
        self.cv2 = Conv(c_, c2, k[1], 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        """'forward()' applies the YOLOv5 FPN to input data."""
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    """CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks."""

    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.SiLU()
        self.m = nn.Sequential(*(Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)))

    def forward(self, x):
        """Applies a CSP bottleneck with 3 convolutions."""
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), 1))))


class Head(nn.Module):
    """classification head, i.e. x(b,c1,20,20) to x(b,c2)."""

    # ch_in, ch_out, num_head, kernel, stride, padding, groups
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1):
        super().__init__()
        c_ = 1280  # efficientnet_b0 size
        self.conv = Conv(c1, c_, k, s, p, g)
        self.pool = nn.AdaptiveAvgPool2d(1)  # to x(b,c_,1,1)
        self.drop = nn.Dropout(p=0.01, inplace=True)
        self.linear = nn.Linear(c_, c2)  # to x(b,c2)

    def forward(self, x):
        """Performs a forward pass of the model on input image data."""
        if isinstance(x, list):
            x = torch.cat(x, 1)
        x = self.linear(self.drop(self.pool(self.conv(x)).flatten(1)))
        return x if self.training else x.softmax(1)



import inspect
import torch.nn as nn


class Loss:
    def __init__(self, config):
        loss_name = config["name"]
        loss_config = config[loss_name]
        self.loss = None
        if hasattr(nn, loss_name):
            loss = getattr(nn, loss_name)
            if isinstance(loss, type):
                sig = inspect.signature(loss.__init__)
                params = {k: v for k, v in loss_config.items() if k in sig.parameters}
                self.loss = loss(**params)

    def __call__(self):
        assert (self.loss is not None)
        return self.loss



from . import augment
from torchvision import transforms


class Transform:
    def __init__(self, config, header):
        augment_list = []
        for augment_name, augment_config in config.items():
            if hasattr(augment, augment_name):
                transfer = getattr(augment, augment_name)
                if isinstance(transfer, type):
                    augment_list.append(transfer(augment_config, header))
        self.transform = transforms.Compose(augment_list)

    def __call__(self, image, label):
        image, label = self.transform((image, label))
        return image, label



# Ultralytics YOLO 🚀, AGPL-3.0 license

import re
from pathlib import Path

import pkg_resources as pkg
from setuptools import find_packages, setup

# Settings
FILE = Path(__file__).resolve()
PARENT = FILE.parent  # root directory
# README = (PARENT / 'README.md').read_text(encoding='utf-8')
REQUIREMENTS = [f'{x.name}{x.specifier}' for x in pkg.parse_requirements((PARENT / 'requirements.txt').read_text())]


# Hereby note to prove that I have been here.
def get_version():
    file = PARENT / 'soar/__init__.py'
    return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', file.read_text(encoding='utf-8'), re.M)[1]

setup(
    name='soarv1',  # name of pypi package
    version=get_version(),  # version of pypi package
    python_requires='>=3.8',
    license='AGPL-3.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows', ]
    )



from torch.utils.data.dataloader import DataLoader
from torch.utils.data import DistributedSampler


class LMDBDataLoader(DataLoader):
    def __init__(self, config, dataset, mode, *args, **kwargs):
        self.mode = mode
        self.config = config[mode]
        self.world_size = config["world_size"]
        self.batch_size = self.config["batch_size"] // self.world_size
        self.num_workers = self.config["num_workers"] // self.world_size
        self.drop_last = self.config["drop_last"]
        self.sampler = LMDBDataSampler(config, dataset, mode)
        super(LMDBDataLoader, self).__init__(
            dataset=dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            sampler=self.sampler,
            drop_last=self.drop_last,
            *args, **kwargs)


class LMDBDataSampler(DistributedSampler):
    def __init__(self, config, dataset, mode):
        self.rank = config["rank"]
        self.world_size = config["world_size"]
        self.config = config[mode]
        self.shuffle = self.config["shuffle"]
        super(LMDBDataSampler, self).__init__(
            dataset=dataset,
            num_replicas=self.world_size,
            rank=self.rank,
            shuffle=self.shuffle)



import os
import lmdb
from .crypt import Crypt


class Storager:
    def __init__(self, save_path, key):
        self.save_path = save_path
        self.lmdbFile = lmdb.open(self.save_path, int(10 * 1024 ** 3))
        self.crypt = Crypt(key)

    def __del__(self):
        self.lmdbFile.close()

    def read(self, key):
        with self.lmdbFile.begin() as context:
            buffer = context.get(self.crypt.encode(key))
        return buffer

    def write(self, key, value):
        with self.lmdbFile.begin(write=True) as context:
            context.put(self.crypt.encode(key), value.getvalue())
